#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vector>
#include <math.h>
#include <algorithm>
#include <time.h>
#include <set>
#include <map>


#include <pthread.h>

#define RIGHT 1
#define LEFT 2
#define ID_TYPE int

using namespace std;
const char ACGT[4] = {'A','C','G','T'};


typedef enum{
    JOB_NEW,
    JOB_DONE,
} jobStatus ;


typedef struct linkingNode{
    ID_TYPE id;
    int len;
}linkingNode;


typedef struct breakNode{
    int breakType;
    int pos;
}breakNode;


typedef struct neuronNode{
    ID_TYPE index;
    ID_TYPE inverseID;
    int readNum;
    char *seq;
    int GLOBAL_visit;
    bool GLOBAL_used;
}neuronNode;


typedef struct pathNode{
    ID_TYPE index;
    char *seq;
}pathNode;


const int KMERSIZE = 15;
const unsigned long MAXINDEX = 1UL<<(KMERSIZE<<1);
ID_TYPE *indexTable;


int MINMATCH,MAXMATCH,OVERLAPLEN,NONoverlap,THREAD_NUM;
const int MINKMER =0;
double MINDENSITY = 0.3;
double BRANCHDENSITY = 0.3;
double CUT_RATIO = 0.25 ;

size_t SHUFFLE_SHIFT = 1;

int seqLen,contigNum;
ID_TYPE seqNum ;
ID_TYPE readTotal;
neuronNode *table;

class usedNode{
private:
    bool value ;

public:
    bool returned_false ;

    usedNode(){
        returned_false = false ;
    }

    bool get(){
        if( !value )
            returned_false = true ;
        return value ;
    }

    void set( bool value ){
        this->value = value ;
    }

    void fetch_false(){
        value = false ;
        returned_false = true ;
    }
};


class visitNode{
public:
    int ref ;
    int value ;

    int &table_fetch(ID_TYPE i){
        ref = table[i].GLOBAL_visit ;
        value = ref ;
        return value ;
    }
};


class bufferNode{
public:
    map <ID_TYPE,set<linkingNode> > output;
    vector <pathNode> path;

    ~bufferNode(){
        for( size_t i=0 ; i<path.size() ; i++ )
            delete [] path[i].seq ;
    }
};


class jobNode{
    map <ID_TYPE,usedNode> used;
    map <ID_TYPE,visitNode> visit;

    jobStatus status ;

public:
    ID_TYPE index;

    map <ID_TYPE,set<linkingNode> > *output;
    vector <pathNode> *path;

    char *contig;
    int contig_totalLen;
    vector <breakNode> breakList;

    jobNode(){
        contig = NULL ;
        output = NULL ;
        path = NULL ;
        status = JOB_NEW ;
    }
    ~jobNode(){
        if( contig )
            delete [] contig ;
    }

    void bind_buffer( bufferNode *buffer ){
        output = &(buffer->output) ;
        path = &(buffer->path) ;
    }

    void unbind_buffer(){
        output = NULL ;
        path = NULL ;
    }

    void clear(){
        status = JOB_NEW ;

        used.clear() ;
        visit.clear() ;

        if( output )
            output->clear() ;

        if( path ){
            for( size_t i=0 ; i<path->size() ; i++ )
                delete [] (*path)[i].seq ;
            path->clear() ;
        }

        breakList.clear() ;
        if( contig ){
            delete [] contig ;
            contig = NULL ;
        }
    }

    jobStatus get_status(){
        return status ;
    }

    jobStatus set_status( jobStatus status ){
        this->status = status ;
        return this->status ;
    }

    bool get_used( ID_TYPE i, bool log=true ){
        return table[i].GLOBAL_used ;
    }

    void set_used(ID_TYPE i, bool v){
        table[i].GLOBAL_used = v ;
    }

    int inc_visit(ID_TYPE i){
        return ++table[i].GLOBAL_visit ;
    }
    
    void print_contigs(){
        if( contig ){
//            if(strlen(contig)>seqLen){
//               printf(">contig%d_%d\n%s\n",contigNum++,strlen(contig),contig);
//            }


/*
            int startP = 0;
            for( int i=0 ; i<(int)breakList.size() ; i++ ){
                char bakch = contig[breakList[i].pos] ;
                contig[breakList[i].pos] = '\0' ;
                if( breakList[i].pos - startP >= 2*seqLen ){
                    if( breakList[i].breakType == LEFT )
                        printf(">contig%d_%d_L\n%s\n", contigNum++, breakList[i].pos-startP, contig+startP) ;
                    else
                        printf(">contig%d_%d_R\n%s\n", contigNum++, breakList[i].pos - startP, contig+startP) ;
                }
                contig[breakList[i].pos] = bakch ;
                startP = breakList[i].pos ;
            }
*/

            int startP = 0;
            for( int i=0 ; i<(int)breakList.size() ; i++ ){
                if( breakList[i].pos - startP > seqLen ){
                    if( breakList[i].breakType == LEFT ){
                        if(breakList[i].pos+15 < contig_totalLen){
                            char bakch = contig[breakList[i].pos+15];
                            contig[breakList[i].pos+15] = '\0';
                            printf(">contig%d_%d_L\n%s\n", contigNum++, breakList[i].pos-startP, contig+startP) ;
                            contig[breakList[i].pos+15]=bakch;
                        }else{
                            char bakch = contig[breakList[i].pos];
                            contig[breakList[i].pos] = '\0';
                            printf(">contig%d_%d_L\n%s\n", contigNum++, breakList[i].pos-startP, contig+startP) ;
                            contig[breakList[i].pos]=bakch;
                        }
                    }else{
                        char bakch = contig[breakList[i].pos] ;
                        contig[breakList[i].pos] = '\0' ;
                        printf(">contig%d_%d_R\n%s\n", contigNum++, breakList[i].pos - startP, contig+startP);
                        contig[breakList[i].pos]=bakch;
                        if(breakList[i].pos - 15 >0 )
                          breakList[i].pos-=15;
                   }
                }
                startP = breakList[i].pos ;
            }

            if( startP + seqLen < contig_totalLen )
                printf(">contig%d_%d_G\n%s\n", contigNum++, contig_totalLen-startP, contig+startP) ;

            fflush(stdout) ;
        }
    }
};


class jobAllocator{
    size_t size ;
    size_t assigned ;
    pthread_mutex_t mutex ;

public:
    jobAllocator( size_t size ){
        this->size = size ;
        assigned = 0 ;
        pthread_mutex_init( &mutex, NULL ) ;
    }
    ~jobAllocator(){
        pthread_mutex_destroy( &mutex ) ;
    }

    size_t fetch_range( size_t *base, size_t *lim, size_t bucket_size ){
        pthread_mutex_lock( &mutex ) ;

        *base = assigned ;
        assigned += bucket_size ;
        if( assigned > size )
            assigned = size ;
        *lim = assigned ;

        pthread_mutex_unlock( &mutex ) ;

        return *lim - *base ;
    }

    size_t increase( size_t inc ){
        size_t base, lim ;
        return fetch_range( &base, &lim, inc ) ;
    }
};


class threadArg{
public:
    vector<void*> argv ;

    void push_back( void *v ){
        argv.push_back( v ) ;
    }
} ;


size_t gcd( size_t a, size_t b ){
    if( a == 0 )
        return b ;

    while( b ){
        size_t r = a % b ;
        a = b ;
        b = r ;
    }
    return a ;
}


unsigned long ch2int(char ch){

    if(ch =='A')
        return 0;
    if(ch =='C')
        return 1;
    if(ch =='G')
        return 2;
//    if(ch =='T')
    return 3;
}


unsigned long firstIndex(char* seq){

    unsigned long index;

    index = 0;
    for(int i=0;i<KMERSIZE;i++){
        index = index << 2;
        index = index | ch2int(seq[i]);
    }
    return index;
}


bool operator<(const linkingNode& left, const linkingNode& right) {
    if(table[left.id].readNum != table[right.id].readNum)
        return table[left.id].readNum > table[right.id].readNum;
    return left.id < right.id;
}


int compare(const void * a, const void * b){
    char *ptra,*ptrb;

//    ptra = ((neuronNode*)a)->seq;
//    ptrb = ((neuronNode*)b)->seq;
    ptra = ((neuronNode*)a)->seq+KMERSIZE;
    ptrb = ((neuronNode*)b)->seq+KMERSIZE;
    while(*ptra !='\0' && *ptrb!='\0'){
        if(*ptra != *ptrb)
            return *ptra - *ptrb;
        ptra++;
        ptrb++;

    }
    return 0;
}


int ACGT2int[85] = {
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,1,0,0,0,2,0,0,0,0,0,0,0,0,
    0,0,0,0,3
};


int complement[85] = {
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,'T',0,'G',0,0,0,'C',0,0,0,0,0,0,0,0,
    0,0,0,0,'A'
};


neuronNode* search(neuronNode *temp){
    unsigned long index;
    ID_TYPE  num;

    index = firstIndex(temp->seq);
    num = indexTable[index+1] - indexTable[index]; 
    if(num)
        return (neuronNode*)bsearch(temp,&(table[indexTable[index]]),num,sizeof(neuronNode),compare);
    else
        return NULL;
}


void reverse(char *str){
    int i,j,len;
    char bak;

    len = strlen(str);
    for(i=0,j=len-1;i<=j;i++,j--){
        bak = str[i];
        str[i] = str[j];
        str[j] = bak;
    }
}


void reverse_complement(char *str){
    int i,j,len;
    char bak;

    len = strlen(str);
    for(i=0,j=len-1;i<=j;i++,j--){
        bak = str[i];
        str[i] = complement[(int)str[j]];
        str[j] = complement[(int)bak];
    }
}


ID_TYPE read_seq(char *fileName){
    FILE *fptr;
    ID_TYPE i,n;


    fptr = fopen(fileName,"r");
//    usedTotal = 0;
    if( fscanf(fptr,"readNum= %d\treadLength= %d\n",&n,&seqLen) );

    char *seq_ptr = NULL ;
    table = new neuronNode[n];
    for(i=0;i<n;i++){
//        table[i].seq = new char[seqLen+1];
        if( (i&((1<<16)-1)) == 0 )
            seq_ptr = new char [(seqLen+1)*(1<<16)] ;
        else
            seq_ptr += seqLen+1 ;

        table[i].seq = seq_ptr ;
        table[i].index = i;
        if( fscanf(fptr,"%d\t%s\t%d\n",&(table[i].readNum),table[i].seq,&(table[i].inverseID)) );
        table[i].GLOBAL_visit = 0;
        table[i].GLOBAL_used = false;
        readTotal += table[i].readNum;
    }
    fclose(fptr);
    return n;
}


void buildIndexTable(){
    unsigned long index;
    unsigned long  acc,i,j;

    indexTable = new ID_TYPE[MAXINDEX+1];
    acc = 0;
    j = 1;
    index = firstIndex(table[0].seq);

    for(i=0;i<MAXINDEX;i++){
        while(i > index && j < (unsigned long) seqNum){
            acc++;
            index = firstIndex(table[j++].seq);        
        }
      //  indexTable[i] = acc;
       indexTable[i] = (i>index)?seqNum:acc;
    }
    indexTable[MAXINDEX] = seqNum;
}


int kmerExist(char *seq , double *ratio, int *gapLen,bool allowUsed, jobNode *job){
    int i,j,last,n,preIndex,dist;
    char bakch;
    neuronNode temp;
    neuronNode *ptr;

    last = strlen(seq) - (seqLen);
    preIndex = *gapLen = 0;
    for(n = 0,i = 0,j=i+seqLen;i<=last;i++,j++){
        bakch = seq[j];
        seq[j] = '\0';
        temp.seq = seq+i;
        ptr = search(&temp);
        seq[j] = bakch;

        if(ptr && (allowUsed || !job->get_used(ptr->index))){
            dist = i -preIndex;
            if(dist > *gapLen)
                *gapLen = dist;
            preIndex = i;
            n++;
        }
    }
    *ratio = (double)n / (last+1);
    return n;
}


void buildForwardConnection(ID_TYPE i, jobNode *job){
    set<linkingNode> &i_output = (*(job->output))[i] ;

    if(i_output.size()>0){
        set<linkingNode>::iterator it,del;
        for(it=i_output.begin();it!=i_output.end();){
            ID_TYPE start = it->id;
            if( job->get_used(start) ){
                del = it;
                it++;
                i_output.erase(del);
            }
            else
                it++;
        }

        return;
    }

    ID_TYPE start,end;
    neuronNode temp;
    linkingNode templink;

    int count = 0;
    for(int len=MAXMATCH;len>=MINMATCH;len-=1){
        temp.seq = table[i].seq+seqLen-len;
        neuronNode *ptrOutput = search(&temp);
        if(ptrOutput){
            for(start=ptrOutput->index;start > 0 && strncmp(table[start].seq,temp.seq,len)==0;start--);
            for(end=ptrOutput->index+1;end<seqNum && strncmp(table[end].seq,temp.seq,len)==0;end++);
            
            if(end - start > 10)
                continue;
                
            for(ID_TYPE j=start+1;j<end;j++){
                if(job->get_used(j))
                    continue;
                if(j!=i){
                    templink.id = j;
                    templink.len = len;
                    i_output.insert(templink);
                }
            }
            count++;
            if(count > 5)
                break;
        }
    }
}


ID_TYPE getInverseID(ID_TYPE id){
    if(table[id].inverseID != -1)
        return table[id].inverseID;
    else{
        fprintf( stderr, "Invalid inverseID\n" );
    }
    return -1;
}


size_t getNext( ID_TYPE start, set<ID_TYPE> *candidates, jobNode *job ){
    candidates->clear() ;
    
    buildForwardConnection(start,job);
    set<linkingNode> &start_output = (*(job->output))[start] ;
    for(set<linkingNode>::iterator itStart = start_output.begin(); itStart != start_output.end(); itStart++){
        buildForwardConnection(itStart->id,job);
        set<linkingNode> &itstart_output = (*(job->output))[itStart->id] ;
        for(set<linkingNode>::iterator itExt = itstart_output.begin(); itExt != itstart_output.end(); itExt++){
            if(job->get_used(itExt->id))
                continue;
/*
         buildForwardConnection(itExt->id,job);        
         set<linkingNode> &itfinal_output = (*(job->output))[itExt->id] ;
         if(itfinal_output.size()==0)
            continue;
*/

            candidates->insert(itExt->id);
        }
    }
    return candidates->size() ;
}


int dist(char *s, char *t,int len,int *breakPoint){
    int i,diff;

    *breakPoint = -1;
    for(i=diff=0;i<len;i++){
        if(s[i]=='\0' || t[i]=='\0')
            break;
        if(s[i]!=t[i]){
            if(*breakPoint < 0)
                *breakPoint = i;
            diff++;
        }
    }
    return diff;
}


int consensus(vector<char*> seq, int n, int index){
    int ACGT[4] = {0,0,0,0};
    int i,max;

    for(i=0;i<n;i++)
        ACGT[ACGT2int[(int)seq[i][index]]]++;
    for(max=0,i=1;i<4;i++){
        if(ACGT[i] > ACGT[max])
            max = i;
    }
    return max;
}


void show_gap_consensus(ID_TYPE s, ID_TYPE t,char *gap, bool *overlap, int *overlaplen, jobNode *job){
//    int i,n,len;
    set<linkingNode>::iterator itS,itT,Send,Tend;
    vector<char*> support;
    char *temp;

    *overlap = false;
    set<linkingNode> &s_output = (*(job->output))[s] ;
    for(itS=s_output.begin();itS!=s_output.end();itS++){
        set<linkingNode> &t_output = (*(job->output))[itS->id] ;
        for(itT=t_output.begin();itT!=t_output.end();itT++){
            if(itT->id == t){
                if(itS->len + itT->len>=seqLen){
                    *overlap = true;
                    *overlaplen = itS->len + itT->len - seqLen;
                    return ;
                }
                else{
                    temp = new char[seqLen+1];
                    strcpy(temp,table[itS->id].seq+itS->len);
                    temp[strlen(temp)-itT->len]='\0';
                    support.push_back(temp);
                }
            }
        }
    }

    size_t n = support.size();
    if(n == 0)
        gap[0] = '\0';
    else{
        size_t len = strlen(support[0]);
        size_t i ;
        for( i=0 ; i<len ; i++ )
            gap[i] = ACGT[consensus(support,n,i)];
        gap[i] = '\0';
        for(i=0;i<n;i++)
            delete[] support[i];
    }
    support.clear();
}


void mark_used(char *seq, jobNode *job){
    int i,last;
    ID_TYPE inverseID;
    neuronNode temp;
    neuronNode *ptr;
    char tempch;

    last = strlen(seq) - seqLen;
    for(i = 0;i<=last;i++){
        temp.seq = seq+i;
        tempch = seq[i+seqLen];
        seq[i+seqLen] = '\0';
        ptr = search(&temp);
        seq[i+seqLen] = tempch;

        if(ptr){
            job->set_used(ptr->index,true);
//            usedTotal += ptr->readNum;
            inverseID = getInverseID(ptr->index);
            job->set_used(inverseID,true);
//            usedTotal += table[inverseID].readNum;
        }
    }
}


bool travel(ID_TYPE,int,char*,double,jobNode*);
int testBranch2(char *seq,int *branchType, int *breakPoint, jobNode *job){
    int depth;
    ID_TYPE pItem,index;
    bool extendable;
    char *backbone;
    neuronNode temp,Rtemp;
    neuronNode *ptrOutput;
    char Rori[1000];

    temp.seq = new char[seqLen+1];
    depth = 2;
    backbone = new char[(seqLen+1)*depth*2];
    Rtemp.seq = new char[seqLen+1];
    strncpy(Rori,seq,seqLen);
    Rori[seqLen] = '\0';
    reverse_complement(Rori);

    strncpy(Rtemp.seq, seq+NONoverlap,OVERLAPLEN);
    Rtemp.seq[OVERLAPLEN] = '\0';
    reverse_complement(Rtemp.seq);
    ptrOutput = search(&Rtemp);
    if(ptrOutput){
        pItem = ptrOutput->index;
        for(index=pItem;index > 0 && strncmp(table[index].seq,Rtemp.seq,OVERLAPLEN)==0;index--){  
            if(!job->get_used(index)){
                memset(backbone,0,sizeof(char)*depth*(seqLen+1));
                extendable = travel(index,1,backbone,BRANCHDENSITY,job);
                char test[1000];
                strcpy(test,seq);
                test[seqLen] = '\0';
                reverse_complement(test);
                if(extendable && dist(backbone,test,seqLen+1,breakPoint)> 0){
                    delete[] temp.seq;
                    delete[] Rtemp.seq;
                    delete[] backbone;
                    *branchType = LEFT;
                    return LEFT;
                }
            }
            else if(dist(table[index].seq,Rori,seqLen+1,breakPoint)> 0){
                delete[] temp.seq;
                delete[] Rtemp.seq;
                delete[] backbone;
                *branchType = LEFT;
                return LEFT;
            }
        }

        for(index=pItem+1; index<seqNum && strncmp(table[index].seq,Rtemp.seq,OVERLAPLEN)==0;index++){  
            if(!job->get_used(index)){
                memset(backbone,0,sizeof(char)*depth*(seqLen+1));
                extendable = travel(index,1,backbone,BRANCHDENSITY,job);
                char test[1000];
                strcpy(test,seq);
                test[seqLen] = '\0';
                reverse_complement(test);
                if(extendable && dist(backbone,test,seqLen+1,breakPoint)> 0){
                    delete[] temp.seq;
                    delete[] Rtemp.seq;
                    delete[] backbone;
                    *branchType = LEFT;
                    return LEFT;
                }
            }
            else if(dist(table[index].seq,Rori,seqLen+1,breakPoint)> 0){
                delete[] temp.seq;
                delete[] Rtemp.seq;
                delete[] backbone;
                *branchType = LEFT;
                return LEFT;
            }
        }
    }

    delete[] Rtemp.seq;
    strncpy(temp.seq,seq,OVERLAPLEN);
    temp.seq[OVERLAPLEN] = '\0';
    ptrOutput = search(&temp);
    if(ptrOutput){
        pItem = ptrOutput->index;
        for(index=pItem;index > 0 && strncmp(table[index].seq,temp.seq,OVERLAPLEN)==0;index--){  
            if( !job->get_used(index) ){
                memset(backbone,0,sizeof(char)*depth*(seqLen+1));
                extendable = travel(index,1,backbone,BRANCHDENSITY,job);
                if (extendable && dist(backbone,seq,seqLen+1,breakPoint)>0){
                    delete[] temp.seq;
                    delete[] backbone;
                    *branchType = RIGHT;
                    return RIGHT;
                }
            }
            else if(dist(table[index].seq,seq,seqLen+1,breakPoint)> 0){
                delete[] temp.seq;
                delete[] backbone;
                *branchType = RIGHT;
                return RIGHT;
            }

        }
        for(index=pItem+1; index<seqNum && strncmp(table[index].seq,temp.seq,OVERLAPLEN)==0;index++){  
            if( !job->get_used(index) ){
                memset(backbone,0,sizeof(char)*depth*(seqLen+1));
                extendable = travel(index,1,backbone,BRANCHDENSITY,job);
                if (extendable && dist(backbone,seq,seqLen+1,breakPoint)>0){
                    delete[] temp.seq;
                    delete[] backbone;
                    *branchType = RIGHT;
                    return RIGHT;
                }
            }
            else if(dist(table[index].seq,seq,seqLen+1,breakPoint)> 0){
                delete[] temp.seq;
                delete[] backbone;
                *branchType = RIGHT;
                return RIGHT;
            }
        }
    }
    delete[] temp.seq;
    delete[] backbone;
    return 0;
}


ID_TYPE back_trimming(ID_TYPE start, jobNode *job){
    bool backcheck;
    neuronNode temp;
    vector<pathNode>::iterator pathit,it;
    char *backup;
    int last;
//    int count;
    neuronNode *ptr=NULL;
    char tempch;

    vector <pathNode> &path = *(job->path);

    pathit=path.end()-1;
    backup = new char[seqLen+1 + strlen(pathit->seq)+1];
    sprintf(backup,"%s%s",table[pathit->index].seq,pathit->seq);
    for(backcheck=false,last =strlen(backup)-1;last > seqLen+1;last--){
//    for(backcheck=false,last =strlen(backup)-1,count=0;count<seqLen/3;count++,last--){
        tempch = backup[last];
        backup[last]='\0';
        temp.seq = backup+(last-seqLen);
        ptr = search(&temp);

        if(ptr && ptr->index != start && travel(ptr->index,1,NULL,MINDENSITY,job)){
            backcheck = true;
            pathit->seq[last-seqLen] = '\0';
            break;
        }
        else if(ptr){
            job->set_used(ptr->index,false);
            job->set_used(getInverseID(ptr->index),false);
        }

    }

    job->set_used(start,false);
    job->set_used(getInverseID(start),false);
    delete[] backup;
    if(backcheck)
        return ptr->index;
    delete[] pathit->seq;
    path.pop_back();
    return -1;
}


bool travel(ID_TYPE start,int step,char* backbone, double minDensity, jobNode *job){
 /*
    step > 0 : test whether start can extend step times
    step < 0 : truly extend
*/

    set<ID_TYPE> NextCandidates;

    char *gap = new char[seqLen+1];
    char *seq = new char[3*(seqLen+1)];
    char *bak = new char[3*(seqLen+1)];

    vector <pathNode> &path = *(job->path);

    while(1){
        if(step < 0){
            mark_used(table[start].seq,job);
            int visit = job->inc_visit(start) ;
            if(visit > 3){
                break;
            }
        }
        if( 0 == getNext(start,&NextCandidates,job) ){
            if(step > 0 || path.size() < 3){
                delete[] gap;
                delete[] seq;
                delete[] bak;
                return false;
            }
            start = back_trimming(start,job);
            if(start > 0){
                continue;
            }
            break;
        }

        ID_TYPE index = -1;
        double MaxRatio = 0;
        int MaxGap = 0;
        for(set<ID_TYPE>::iterator it = NextCandidates.begin();it!= NextCandidates.end();it++){
            ID_TYPE next = *it;
            bool overlap ;
            int overlaplen ;
            show_gap_consensus(start,next,gap,&overlap,&overlaplen,job);
            if(!overlap){
                sprintf(seq,"%s%s%s",table[start].seq,gap,table[next].seq);
            }
            else{
                sprintf(seq,"%s%s",table[start].seq,table[next].seq+overlaplen);
            }

            double ratio ;
            int gapLen ;
            kmerExist(seq,&ratio,&gapLen,false,job);

// set<ID_TYPE> testCandidates;
            if(ratio > MaxRatio){ // && getNext(next,&testCandidates,job)>0){ 
                index = next;
                MaxRatio = ratio;
                MaxGap = gapLen;
                strcpy(bak,seq);
            }
        }

        if (index != -1 && MaxRatio < minDensity){
            if( step < 0){
                char test[1000];
                test[0] = '\0';
                travel(table[index].inverseID,1,test,0,job);
                if(test[0]){
                    reverse_complement(test);

                    double ratio ;
                    int gapLen ;
                    kmerExist(test,&ratio,&gapLen,false,job);
                    if(ratio > minDensity)
                        index = -1;
                }
            }
            else
                index = -1;
        }
        NextCandidates.clear();
        if(index == -1 ){
            if(step < 0 && path.size() > 3){
                start = back_trimming(start,job);
                if(start > 0)
                    continue;
            }
            break;
        }
        if(step > 0){  // test mode
            start = index;
            if(backbone != NULL){
                strcpy(backbone,bak);
            }
            step--;
            if(step == 0)
                break;
            continue;
        }

        mark_used(bak,job);

        pathNode ptemp;
        ptemp.seq = new char[2*(seqLen+1)];
        sprintf(ptemp.seq,"%s",bak+seqLen);
        ptemp.index = start;
        path.push_back(ptemp);

        if(index > 0){
            start = index;
        }
        else
            break;
    }

    delete[] gap;
    delete[] seq;
    delete[] bak;
    if(step == 0)
        return true;
    return false;
}


void free_tables(){
    for( ID_TYPE i=0 ; i<seqNum ; i+=(1<<16) ){
        delete[] table[i].seq;
    }

    delete[] table;
}


int read_distribution_cut(double cut){
    int binsize,index,threshold;
    ID_TYPE i;
    set<int>bin;

    for(i=0;i<seqNum;i++)
        bin.insert(table[i].readNum);
    binsize = bin.size();
    index = 0;
    threshold = (int) (binsize * cut);
    for(set<int>::iterator it=bin.begin();it!=bin.end();it++,index++){
//        if(index == threshold)        // original code
        if(index >= threshold)          // modified 2013.06.05
            return *it;
    }

//    return threshold;         // original codes
    return *bin.rbegin() ;      // modified 2013.06.05
}


bool goodSeed_pre(ID_TYPE start, int seedCut){
    if( table[start].GLOBAL_used )
        return false ;

    int FreadNum = table[start].readNum;
    int RreadNum = table[table[start].inverseID].readNum;
    if((FreadNum > 1 && FreadNum <= seedCut) || (RreadNum > 1 && RreadNum <= seedCut))
        return true ;

    return false ;
}


bool goodSeed(ID_TYPE start, jobNode *job){
    if( job->get_used(start) ){        
//        fprintf( stderr, "%d: Bad Seed by USED!\n", start ) ;
        return false;
    }

    buildForwardConnection(start,job);
    ID_TYPE inverseID = getInverseID(start);
    buildForwardConnection(inverseID,job);
    if( travel(inverseID,1,NULL,MINDENSITY,job) && travel(start,1,NULL,MINDENSITY,job)){
//        fprintf( stderr, "%d: Good Seed.\n", start ) ;
        return true;
    }

//    fprintf( stderr, "%d: Bad Seed by TEST.\n", start ) ;
    return false;
}


bool doJR_job( jobNode *job ){
    ID_TYPE start = job->index;
    job->clear() ;

    if(!goodSeed(start,job))
        return false ;

    vector <pathNode> &path = *(job->path);

    pathNode ptemp;
    ptemp.index = start;
    ptemp.seq = new char[seqLen+1];
    strcpy(ptemp.seq,table[start].seq);
    reverse_complement(ptemp.seq);
    path.push_back(ptemp);

    travel(getInverseID(start),-1,NULL,MINDENSITY,job);

    for(int i=0;i<(int)path.size();i++)
        reverse_complement(path[i].seq);
    std::reverse(path.begin(),path.end());
    travel(start,-1,NULL,MINDENSITY,job);

    if(path.size()>1){
        vector<pathNode>::iterator it;
        int &totalLen = job->contig_totalLen ;
        totalLen = 0 ;
        for( it = path.begin(); it != path.end(); it++){
            totalLen+=strlen(it->seq);
        }
        char *contig = job->contig = new char[totalLen+1];

        strcpy(contig,path.begin()->seq);
        for(it=path.begin()+1;it!=path.end();it++)
            strcat(contig,it->seq);

        vector <breakNode> &breakList = job->breakList;

        char test[1000];
        for(int i=0;i<totalLen-seqLen-1;i++){
            strncpy(test,contig+i,seqLen);
            test[seqLen] = '\0';
            int branchType = 0;
            int breakPoint;

            testBranch2(test,&branchType,&breakPoint,job);
            if(branchType){
                breakNode btemp;
                btemp.breakType = branchType;
                btemp.pos = (branchType==RIGHT)?i+breakPoint: i+ seqLen-breakPoint;
                breakList.push_back(btemp);
                i = btemp.pos;
            }
        }

        for(int i =1;i <(int) breakList.size();i++){
            if(breakList[i-1].breakType==LEFT && breakList[i].breakType ==RIGHT ){
                int endP = breakList[i].pos;
                int startP = (breakList[i-1].pos +1 );
                if(endP - startP < seqLen-4 ){

                    char tempSeq[1000];
                    int gapLen;
                    double ratio;
                    strncpy(tempSeq,contig+startP-(seqLen-1-(endP-startP)),seqLen+(seqLen-2-(endP-startP+1)));
                    tempSeq[seqLen+(seqLen-2-(endP-startP+1))] = '\0';
                    kmerExist(tempSeq,&ratio,&gapLen,true,job);
                    if(ratio>MINDENSITY){
                        breakList.erase(breakList.begin()+(i-1),breakList.begin()+(i+1));
                        i-=2;
                    }
                }
            }
        }

        job->set_used(start,true);
        job->set_used(getInverseID(start),true);
    }

    return true ;
}


void *doJR_thread( void *arg ){
    vector<void*> &argv = ((threadArg*)arg)->argv ;
    vector<jobNode*> &job_queue = *(vector<jobNode*>*)argv[0] ;
    size_t bucket_size = *(size_t*)argv[1] ;
    jobAllocator &JR_job = *(jobAllocator*)argv[2] ;
    pthread_mutex_t *write_back_mutex = (pthread_mutex_t*)argv[3] ;
    size_t &write_back_base = *(size_t*)argv[4] ;
    size_t &extended_seeds = *(size_t*)&argv[5] ;

    bufferNode buffer ;

    size_t range_base, range_lim ;
    while( JR_job.fetch_range(&range_base, &range_lim, bucket_size) ){
        for( size_t i_job=range_base ; i_job<range_lim ; i_job++ ){
            jobNode *job = job_queue[i_job] ;
            job->bind_buffer( &buffer ) ;

            if( doJR_job(job) )
                extended_seeds ++ ;

            job->set_status(JOB_DONE) ;
            job->unbind_buffer() ;

            pthread_mutex_lock( write_back_mutex ) ;
            for( ; write_back_base < job_queue.size() ; write_back_base++ ){
                jobNode *job = job_queue[write_back_base] ;
                jobStatus job_status = job->get_status() ;
                if( job_status == JOB_NEW )
                    break ;

                if( job->contig )
                    job->print_contigs() ;

                delete job ;
            }
            pthread_mutex_unlock( write_back_mutex ) ;
        }
    }

    return NULL ;
}


void doJR_queue( int seedCut, size_t range_start, size_t range_size, size_t bucket_size ){
    if( range_start >= (size_t)seqNum )
        return ;
    if( range_start+range_size > (size_t)seqNum )
        range_size = (size_t)seqNum-range_start ;

    vector<jobNode*> job_queue ;
    for( size_t i=0 ; i<range_size ; i++ ){
        ID_TYPE start = range_start + i ;
        start = (SHUFFLE_SHIFT*(size_t)start)%seqNum ;

        if( !goodSeed_pre(start,seedCut) )
            continue;

        job_queue.push_back( new jobNode ) ;
        job_queue.back()->index = start ;
    }
    fprintf( stderr, "%10lu Unused Seeds Selected by Read Distribution.\n", job_queue.size() ) ;

    threadArg arg ;
    arg.push_back( (void*)&job_queue ) ;
    arg.push_back( (void*)&bucket_size ) ;

    jobAllocator JR_job( job_queue.size() ) ;
    arg.push_back( (void*)&JR_job ) ;

    pthread_mutex_t write_back_mutex ;
    pthread_mutex_init( &write_back_mutex, NULL ) ;
    arg.push_back( (void*)&write_back_mutex ) ;

    size_t write_back_base = 0 ;
    arg.push_back( (void*)&write_back_base ) ;
    
    size_t extended_seeds = 0 ;
    arg.push_back( (void*)0 ) ;


    threadArg *arg_array = new threadArg[THREAD_NUM] ;
    pthread_t *thread = new pthread_t[THREAD_NUM] ;
    for( int i=1 ; i<THREAD_NUM ; i++ ){
        arg_array[i] = arg ;
        pthread_create( &thread[i], NULL, doJR_thread, (void*)&(arg_array[i]) ) ;
    }

    doJR_thread( (void*)&arg ) ;
    extended_seeds += (size_t)arg.argv[5] ;

    for( int i=1 ; i<THREAD_NUM ; i++ ){
        pthread_join( thread[i], NULL ) ;
        extended_seeds += (size_t)arg_array[i].argv[5] ;
    }
    delete [] thread ;
    delete [] arg_array ;

    pthread_mutex_destroy( &write_back_mutex ) ;

    fprintf( stderr, "%10lu Extended Seeds.\n", extended_seeds ) ;
}


void generate_shuffle_shift(){
    for( SHUFFLE_SHIFT=1+(size_t)seqNum*7/53 ; gcd(seqNum,SHUFFLE_SHIFT) > 1 ; SHUFFLE_SHIFT++ ) ;
}


int main(int argc, char* argv[]){
    if(argc < 6){
        fprintf( stderr, "./JR <FileName> <minOverlap> <maxOverlap> <ratio> <threadNum>\n" );
        return 0;
    }
    seqNum = read_seq(argv[1]);
    buildIndexTable();

    MINMATCH = atoi(argv[2]);
    MAXMATCH = atoi(argv[3]);
    MINDENSITY = BRANCHDENSITY = atof(argv[4]);

    OVERLAPLEN = (int) (seqLen * 0.9);
    NONoverlap = seqLen-OVERLAPLEN;

    THREAD_NUM = atoi(argv[5]) ;
    if( THREAD_NUM < 1 )
        THREAD_NUM = 1 ;

    contigNum = 1;

    char logName[100];
    sprintf(logName,"JR_%d_%d_%.1lf.log",MINMATCH,MAXMATCH,MINDENSITY);

    int seedCut=read_distribution_cut(CUT_RATIO);
    fprintf( stderr, "seedCut Depth: %d\n", seedCut ) ;
    generate_shuffle_shift() ;

    size_t base = 0 ;
    size_t size = (size_t)(seqNum*0.0002) ;
    if( size < 1000 )
        size = 1000 ;

    fprintf( stderr, "STAGE 1:\n" ) ;
    doJR_queue( seedCut, base, size, 1 ) ;
    base += size ;

    size = (size_t)(seqNum*0.001) ;
    for( int i=2 ; base<(size_t)seqNum ; i++ ){
        fprintf( stderr, "STAGE %d:\n", i ) ;
        doJR_queue( seedCut, base, size, 5 ) ;
        base += size ;
        size *= 2 ;
    }

    free_tables();
    return 0;
}



