#include <stdio.h>
#include <vector>
#include <string.h>
#include <unistd.h>

using namespace std;

typedef struct pairNode{
    char read1[1000];
    char read2[1000];
    int insertSize;
    int length;
}pairNode;


bool runErrorCorrection=false,runFLASH=false,runContigOrder=false,runGapFilling=false;
int  trim5,readLen,resampleShift,minOverlap,maxOverlap,minContigLen,nThread;
double ratio;
char JR_Assembler_BIN[1000];
char SSPACE_PATH[1000];
char FLASH_PATH[1000];
char SOAPec_DIR[1000];
char KmerFreq_PATH[1000];
char Corrector_PATH[1000];
char merge_pair_list_PATH[1000];
char Project_HOME_DIR[1000];
char WORKSPACE_DIR[1000];
char PJname[1000];
vector<pairNode> single,PE,MP,frag;



void generate_libraries(){
  FILE *fptr;
  int count = 1;
  char fileName[100];




  sprintf(fileName,"%s.lib",PJname);
  fptr=fopen(fileName,"w");
  for(int i=0;i<frag.size();i++){
   if(frag[i].length > 45)
      fprintf(fptr,"lib%d bowtie %s_45len %s_45len %d 0.75 FR\n",count++,frag[i].read1,frag[i].read2,frag[i].insertSize);
   else
      fprintf(fptr,"lib%d bowtie %s %s %d 0.75 FR\n",count++,frag[i].read1,frag[i].read2,frag[i].insertSize);
  }
  for(int i=0;i<PE.size();i++){
   if(PE[i].length > 45)
       fprintf(fptr,"lib%d bowtie %s_45len %s_45len %d 0.75 FR\n",count++,PE[i].read1,PE[i].read2,PE[i].insertSize);
   else
       fprintf(fptr,"lib%d bowtie %s %s %d 0.75 FR\n",count++,PE[i].read1,PE[i].read2,PE[i].insertSize);
  }
  for(int i=0;i<MP.size();i++){
   if(MP[i].length > 45)
       fprintf(fptr,"lib%d bowtie %s_45len %s_45len %d 0.75 RF\n",count++,MP[i].read1,MP[i].read2,MP[i].insertSize);
   else
       fprintf(fptr,"lib%d bowtie %s %s %d 0.75 RF\n",count++,MP[i].read1,MP[i].read2,MP[i].insertSize);
  }
  fclose(fptr);

}

void generate_script(char *script_file){
  FILE *fptr;
  

   fptr = fopen(script_file,"w");

   sprintf(WORKSPACE_DIR,"%s/WORKSPACE",Project_HOME_DIR);
   fprintf(fptr,"if [ ! -d %s ]; then\n",WORKSPACE_DIR);
   fprintf(fptr,"mkdir %s %s/output\n",WORKSPACE_DIR,WORKSPACE_DIR,WORKSPACE_DIR);
   fprintf(fptr,"fi\n");



     if(runErrorCorrection){
       FILE *flist;
       char filelist[1000];

       
       fprintf(fptr,"echo \"Running ErrorrCorrection\"\n");
       sprintf(filelist,"%s/filelist.txt",Project_HOME_DIR);
       flist = fopen(filelist,"w");
       for(int i=0;i<single.size();i++)
         fprintf(flist,"%s\n",single[i].read1);
       for(int i=0;i<frag.size();i++)
         fprintf(flist,"%s\n%s\n",frag[i].read1,frag[i].read2);
       for(int i=0;i<PE.size();i++)
         fprintf(flist,"%s\n%s\n",PE[i].read1,PE[i].read2);
       for(int i=0;i<MP.size();i++)
         fprintf(flist,"%s\n%s\n",MP[i].read1,MP[i].read2);      
       fclose(flist);
       fprintf(fptr,"%s -k 17 -t %d -p %s/%s %s/filelist.txt \n",KmerFreq_PATH,nThread,WORKSPACE_DIR,PJname,Project_HOME_DIR);
       fprintf(fptr,"%s -k 17 -l 3 -r 50 -t %d -o 3 %s/%s.freq.cz %s/%s.freq.cz.len %s/filelist.txt\n",Corrector_PATH,nThread,WORKSPACE_DIR,PJname,WORKSPACE_DIR,PJname,Project_HOME_DIR);
       
       fprintf(fptr,"cat ");
       for(int i=0;i<single.size();i++)
         fprintf(fptr,"%s.cor.single.fq ",single[i].read1);
       for(int i=0;i<frag.size();i++)
         fprintf(fptr,"%s.cor.pair_1.fq %s.cor.pair_2.fq %s.cor.single.fq %s.cor.single.fq ",frag[i].read1,frag[i].read2,frag[i].read1,frag[i].read2);
       for(int i=0;i<PE.size();i++)
         fprintf(fptr,"%s.cor.pair_1.fq %s.cor.pair_2.fq %s.cor.single.fq %s.cor.single.fq ",PE[i].read1,PE[i].read2,PE[i].read1,PE[i].read2);
       for(int i=0;i<MP.size();i++)
         fprintf(fptr,"%s.cor.pair_1.fq %s.cor.pair_2.fq %s.cor.single.fq %s.cor.single.fq ",MP[i].read1,MP[i].read2,MP[i].read1,MP[i].read2);
       fprintf(fptr,"| %s/toFasta | %s/filterN 0 | %s/trimReadKmer %d 0 %d 0 | %s/buildTable > %s/%s.his\n",JR_Assembler_BIN,JR_Assembler_BIN,JR_Assembler_BIN,trim5,readLen,JR_Assembler_BIN,WORKSPACE_DIR,PJname);
       fprintf(fptr,"rm filelist.* %s/%s.genome_estimate %s/%s.freq.*\n",WORKSPACE_DIR,PJname,WORKSPACE_DIR,PJname);
     }
      if(runFLASH){
       fprintf(fptr,"echo \"Running FLASH\"\n");
       for(int i=0;i<frag.size();i++)
         fprintf(fptr,"%s %s %s -d %s/overlap -o frag%d\n",FLASH_PATH,frag[i].read1,frag[i].read2,WORKSPACE_DIR,i+1);

      }
      fprintf(fptr,"echo \"Building read table\"\n");
      fprintf(fptr,"cat ");
       for(int i=0;i<single.size();i++)
         fprintf(fptr,"%s ",single[i].read1);
       if(!runFLASH){
         for(int i=0;i<frag.size();i++)
           fprintf(fptr,"%s %s ",frag[i].read1,frag[i].read2);
       } 
       for(int i=0;i<PE.size();i++)
         fprintf(fptr,"%s %s ",PE[i].read1,PE[i].read2);
       for(int i=0;i<MP.size();i++)
         fprintf(fptr,"%s %s ",MP[i].read1,MP[i].read2);

      if(runFLASH){
        fprintf(fptr,"%s/overlap/*.extendedFrags.fastq ",WORKSPACE_DIR);
        fprintf(fptr,"| %s/toFasta | %s/filterN 0 | %s/trimReadKmer %d 0 %d %d | %s/buildTable > %s/%s.his\n",JR_Assembler_BIN,JR_Assembler_BIN,JR_Assembler_BIN,trim5,readLen,resampleShift,JR_Assembler_BIN,WORKSPACE_DIR,PJname);
      }else
         fprintf(fptr,"| %s/toFasta | %s/filterN 0 | %s/trimReadKmer %d 0 %d 0 | %s/buildTable > %s/%s.his\n",JR_Assembler_BIN,JR_Assembler_BIN,JR_Assembler_BIN,trim5,readLen,JR_Assembler_BIN,WORKSPACE_DIR,PJname);

    fprintf(fptr,"%s/JR %s/%s.his %d %d %lf %d > %s/%s.initial.fasta\n",JR_Assembler_BIN,WORKSPACE_DIR,PJname,minOverlap,maxOverlap,ratio,nThread,WORKSPACE_DIR,PJname);

  if(single.size()>0 && frag.size()==0 && PE.size()==0 && MP.size()==0){      
     fprintf(fptr,"%s/JRoverlapcloser %s/%s.initial.fasta %s/%s.his 30 > %s/output/%s.final.contigs\n",JR_Assembler_BIN,WORKSPACE_DIR,PJname,WORKSPACE_DIR,PJname,WORKSPACE_DIR,PJname);

   }
  if(runContigOrder &&(frag.size()>0 || PE.size()>0 || MP.size()>0)){
     for(int i=0;i<frag.size();i++){
        if(frag[i].length > 45){
           fprintf(fptr,"cat %s | %s/trimReadKmer 0 0 45 0 > %s_45len\n",frag[i].read1,JR_Assembler_BIN,frag[i].read1);
           fprintf(fptr,"cat %s | %s/trimReadKmer 0 0 45 0 > %s_45len\n",frag[i].read2,JR_Assembler_BIN,frag[i].read2);
        }

     }
     for(int i=0;i<PE.size();i++){
        if(PE[i].length > 45){
           fprintf(fptr,"cat %s | %s/trimReadKmer 0 0 45 0 > %s_45len\n",PE[i].read1,JR_Assembler_BIN,PE[i].read1);
           fprintf(fptr,"cat %s | %s/trimReadKmer 0 0 45 0 > %s_45len\n",PE[i].read2,JR_Assembler_BIN,PE[i].read2);
        }
     }
     for(int i=0;i<MP.size();i++){
        if(MP[i].length > 45){
           fprintf(fptr,"cat %s | %s/trimReadKmer 0 0 45 0 > %s_45len\n",MP[i].read1,JR_Assembler_BIN,MP[i].read1);
           fprintf(fptr,"cat %s | %s/trimReadKmer 0 0 45 0 > %s_45len\n",MP[i].read2,JR_Assembler_BIN,MP[i].read2);
        }
     }
     generate_libraries();
     fprintf(fptr,"perl %s -l %s.lib -s %s/%s.initial.fasta -x 0 -k 5 -b %s\n",SSPACE_PATH,PJname,WORKSPACE_DIR,PJname,PJname);
     fprintf(fptr,"%s/contigLenFilter %s/%s.initial.fasta 0 500 > %s/%s.short.fasta\n",JR_Assembler_BIN,WORKSPACE_DIR,PJname,WORKSPACE_DIR,PJname);
     fprintf(fptr,"%s/getUnused %s/%s.initial.fasta %s/%s.his > %s/unused.fasta\n",JR_Assembler_BIN,WORKSPACE_DIR,PJname,WORKSPACE_DIR,PJname,WORKSPACE_DIR);
     fprintf(fptr,"cat %s/%s.short.fasta %s/unused.fasta > %s/gap.fasta\n",WORKSPACE_DIR,PJname,WORKSPACE_DIR,WORKSPACE_DIR);
     fprintf(fptr,"%s/JRgapcloser %s/%s.initial.fasta %s/gap.fasta %s/%s.final.evidence > %s/%s.x0k5.contigs.fasta\n",JR_Assembler_BIN,WORKSPACE_DIR,PJname,WORKSPACE_DIR,PJname,PJname,WORKSPACE_DIR,PJname);
     fprintf(fptr,"%s/trimContigEnd %s/%s.x0k5.contigs.fasta 5 > %s/%s.x0k5_trim5.contigs.fasta\n",JR_Assembler_BIN,WORKSPACE_DIR,PJname,WORKSPACE_DIR,PJname);
     fprintf(fptr,"perl %s -l %s.lib -s %s/%s.x0k5_trim5.contigs.fasta -x 0 -k 3 -b %s\n",SSPACE_PATH,PJname,WORKSPACE_DIR,PJname,PJname);
     fprintf(fptr,"%s/JRgapcloser %s/%s.x0k5_trim5.contigs.fasta %s/gap.fasta %s/%s.final.evidence > %s.final.all.contigs.fasta\n",JR_Assembler_BIN,WORKSPACE_DIR,PJname,WORKSPACE_DIR,PJname,PJname,PJname);
     fprintf(fptr,"perl %s -l %s.lib -s %s.final.all.contigs.fasta -x 0 -b %s\n",SSPACE_PATH,PJname,PJname,PJname);
     fprintf(fptr,"%s/contigLenFilter %s.final.all.contigs.fasta %d 2000000000 > %s/%s.final.contigs.fasta\n",JR_Assembler_BIN,PJname,minContigLen,PJname,PJname);
     fprintf(fptr,"rm -r %s/alignoutput %s/intermediate_results %s/pairinfo %s/reads\n",PJname,PJname,PJname,PJname);
     fprintf(fptr,"rm %s.final.all.contigs.fasta %s/%s.*.txt %s/%s.final.evidence %s/*.fasta\n",PJname,PJname,PJname,PJname,PJname,WORKSPACE_DIR);
     fprintf(fptr,"mv %s/*final*.fasta %s/output\n",PJname,WORKSPACE_DIR);
     for(int i=0;i<frag.size();i++){
        if(frag[i].length > 45){
           fprintf(fptr,"rm %s_45len %s_45len\n",frag[i].read1,frag[i].read2);
        }
     }
     for(int i=0;i<PE.size();i++){
        if(PE[i].length > 45){
           fprintf(fptr,"rm %s_45len %s_45len\n",PE[i].read1,PE[i].read2);
        }
     }
     for(int i=0;i<MP.size();i++){
        if(MP[i].length > 45){
           fprintf(fptr,"rm %s_45len %s_45len\n",MP[i].read1,MP[i].read2);
        }
     }
   }
   fclose(fptr);


}

int main(int argc, char* argv[]){
   FILE *fptr;
   char line[1000];
   int  n;
   char *chptr;   
   pairNode temp;


   if(argc != 3){
    printf("JR_script configure.txt output.sh\n");
    return 0;
   }
   fptr = fopen(argv[1],"r");
   if(!fptr){
     printf("%s open error!\n",argv[1]);
     return 0;
   }

   while(fgets(line,sizeof(line)-1,fptr)!=NULL){
      if(line[0] == '#' || line[0]=='\n')
           continue;
      if(strstr(line,"JR_Assembler_BIN="))
        sscanf(line,"JR_Assembler_BIN= %s",JR_Assembler_BIN);
     
      if(strstr(line,"SSPACE_PATH="))
        sscanf(line,"SSPACE_PATH= %s",SSPACE_PATH);      
      if(strstr(line,"FLASH_PATH="))
        sscanf(line,"FLASH_PATH= %s",FLASH_PATH);
      if(strstr(line,"SOAPec_DIR=")){
        sscanf(line,"SOAPec_DIR= %s",SOAPec_DIR);
        if(SOAPec_DIR[strlen(SOAPec_DIR)-1]=='/')
          SOAPec_DIR[strlen(SOAPec_DIR)-1]= '\0';
      }
      if(strstr(line,"Project_HOME_DIR=")){
        sscanf(line,"Project_HOME_DIR= %s",Project_HOME_DIR);
        if(Project_HOME_DIR[strlen(Project_HOME_DIR)-1]=='/')
          Project_HOME_DIR[strlen(Project_HOME_DIR)-1] = '\0';
      }
      if(strstr(line,"Project name="))
         sscanf(line,"Project name= %s\n",PJname);
      if(strstr(line,"Number of single-end libraries=")){ 
          sscanf(line,"Number of single-end libraries= %d",&n);
          for(int i=0;i<n;i++){
            while(1){
               fgets(line,sizeof(line)-1,fptr);
               if(line[0]!='#')
                 break;
            } 
            sscanf(line,"%s %d\n",temp.read1,&(temp.length));
            single.push_back(temp);
          }
      }
      if(strstr(line,"Number of overlapping paired-end libraries=")){
          sscanf(line,"Number of overlapping paired-end libraries= %d",&n);
          for(int i=0;i<n;i++){
            while(1){
               fgets(line,sizeof(line)-1,fptr);
               if(line[0]!='#')
                 break;
            } 
            sscanf(line,"%s %s %d %d\n",temp.read1,temp.read2,&(temp.insertSize),&(temp.length));
            frag.push_back(temp);
          }
      }
      if(strstr(line,"Number of non-overlapping paired-end libraries=")){
          sscanf(line,"Number of non-overlapping paired-end libraries= %d",&n);
          for(int i=0;i<n;i++){
            while(1){
               fgets(line,sizeof(line)-1,fptr);
               if(line[0]!='#')
                 break;
            }

            sscanf(line,"%s %s %d %d\n",temp.read1,temp.read2,&(temp.insertSize),&(temp.length));
            PE.push_back(temp);
          }
      }
      if(strstr(line,"Number of mate-pair libraries=")){
          sscanf(line,"Number of mate-pair libraries= %d",&n);
          for(int i=0;i<n;i++){
            while(1){
               fgets(line,sizeof(line)-1,fptr);
               if(line[0]!='#')
                 break;
            }
            sscanf(line,"%s %s %d %d\n",temp.read1,temp.read2, &(temp.insertSize),&(temp.length));
            MP.push_back(temp);
          }
      }
  //    if(strstr(line,"Low-complexity region removal") && strstr(line,"Yes"))
  //      runMdust = true;
      if(strstr(line,"Number of bases at 5' end to be trimmed="))
         sscanf(line,"Number of bases at 5' end to be trimmed= %d",&trim5);
      if(strstr(line,"Assembly Read Length="))
         sscanf(line,"Assembly Read Length= %d",&readLen);
      if(strstr(line,"Base correction=") && strstr(line,"Yes"))
         runErrorCorrection = true;
      if(frag.size()>0 && strstr(line,"Overlapping paired read connection") && strstr(line,"Yes"))
         runFLASH = true;
      if(runFLASH && strstr(line,"Resampling read shift="))
         sscanf(line,"Resampling read shift= %d",&resampleShift);
      if(strstr(line,"Minimum overlap length="))
         sscanf(line,"Minimum overlap length= %d",&minOverlap);
      if(strstr(line,"Maximum overlap length="))
         sscanf(line,"Maximum overlap length= %d",&maxOverlap);
      if(strstr(line,"Minimum remapping ratio="))
         sscanf(line,"Minimum remapping ratio= %lf",&ratio);
      if(strstr(line,"Minimum contig length="))
         sscanf(line,"Minimum contig length= %d",&minContigLen);
      if(strstr(line,"Number of thread="))
         sscanf(line,"Number of thread= %d",&nThread);
      if(strstr(line,"Contig order construction=") && strstr(line,"Yes"))
          runContigOrder = true;     
      if(strstr(line,"Small gap filling=") && strstr(line,"Yes"))
          runGapFilling = true;
   }

   bool fail = false;
   if( access(SSPACE_PATH,R_OK)!=0){
     printf("SSPACE is required but not found!!\n");
     fail = true;
   }
   if( access(FLASH_PATH,X_OK)!=0){
     printf("FLASH is required but not found!!\n");
     fail = true;
   }
//   if(runErrorCorrection ){
    sprintf(KmerFreq_PATH,"%s/bin/KmerFreq_AR",SOAPec_DIR);
    sprintf(Corrector_PATH,"%s/bin/Corrector_AR",SOAPec_DIR);
    if(access(KmerFreq_PATH,X_OK)!=0 || access(Corrector_PATH,X_OK)!=0 ){
         printf("SOAPec is required but not found\n");
         fail = true;
    }
    char JR_exe[1000];

    sprintf(JR_exe,"%s/JR",JR_Assembler_BIN);
    if(access(JR_exe,X_OK)!=0){
       printf("JR bin director is not found\n");
       fail = true;
    }


//   }
  if(fail)
     return 0;
  
   printf("Summary of Project %s\n",PJname);
   printf("-------------------------------------------------------------------\n");
   printf("A. Raw read information\n");
   for(int i=0;i<single.size();i++)
     printf(" %s\n",single[i].read1);
   for(int i=0;i<frag.size();i++)
     printf(" %s %s\n",frag[i].read1,frag[i].read2);
   for(int i=0;i<PE.size();i++)
     printf(" %s %s\n",PE[i].read1,PE[i].read2);
   for(int i=0;i<MP.size();i++)
     printf(" %s %s\n",MP[i].read1,MP[i].read2);
   printf("B. Raw read data processing\n");
 //  if(runMdust)
 //    printf(" Low-complexity region removing= Yes\n");
   printf(" Trim #bases from 5' end= %d\n",trim5);
   printf(" Minimum Read Length= %d\n",readLen);
   if(runErrorCorrection)
     printf(" Base correction= Yes\n");
   if(runFLASH){
     printf(" Overlapping paired read connection= Yes\n");
     printf(" Resampling read shift= %d\n",resampleShift);
   }
   printf("C. JR-Assembler parameters\n");
   printf(" Minimum overlap length= %d\n",minOverlap);
   printf(" Maximum overlap length= %d\n",maxOverlap);
   printf(" Minimum remapping ratio= %lf\n",ratio);
   printf(" Number of thread= %d\n",nThread);
   printf("D. Contig order construction\n");
   if(runContigOrder)
      printf(" Contig order construction= Yes\n");
   if(runGapFilling)
      printf(" Samll gap filling= Yes\n");
   printf("-------------------------------------------------------------------\n");
   generate_script(argv[2]);
  
   fclose(fptr);
   return 0;
}
