#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <string>
#include <vector>
#include <math.h>
#include <algorithm>
#include <time.h>
#include <set>


using namespace std;
#define MATCHLEN 15
typedef struct contigNode{
    int len;
    char *seq;
    char *bak;
}contigNode;


vector<contigNode> contigs;
vector<contigNode> gaps;


const char ACGT[4] = {'A','C','G','T'};
char complement(char ch){

     if(ch == 'A')
        return 'T';
     if(ch == 'C')
        return 'G';
     if(ch == 'G')
        return 'C';
     if(ch == 'T')
        return 'A';
    if(ch == 'a')
        return 't';
     if(ch == 'c')
        return 'g';
     if(ch == 'g')
        return 'c';
     if(ch == 't')
        return 'a';
     return ch;
}
void reverse_complement(char *str){
   int i,j,len;
   char bak;

   len = strlen(str);
   for(i=0,j=len-1;i<=j;i++,j--){
     bak = str[i];
     str[i] = complement(str[j]);
     str[j] = complement(bak);
   }
}



void read_gaps(char* fileName){
  char line[10000];
  contigNode temp;
  string tempSeq;
  int len;
  char *ch;
  FILE *fptr;

   fptr = fopen(fileName,"r");
  if(fptr==NULL)
     return;

  len = 0;
  tempSeq.clear();
  while(fgets(line,sizeof(line)-1,fptr)!=NULL){
         if(line[0]=='>' && len>0){
           temp.seq = new char[tempSeq.size()+1];
           strcpy(temp.seq,tempSeq.c_str());
           temp.bak = new char[tempSeq.size()+1];
           strcpy(temp.bak,tempSeq.c_str());
           temp.len = len;
           gaps.push_back(temp);
           temp.seq = new char[temp.len+1];
           strcpy(temp.seq,gaps[gaps.size()-1].seq);
           reverse_complement(temp.seq);
           gaps.push_back(temp);
           len = 0;
           tempSeq.clear();
         }
         if(line[0]=='>'){
           len = 0;
           tempSeq.clear();
           continue;
         }
         ch = strchr(line,'\n');
         if(ch)
            *ch = '\0';
         len += strlen(line);
         tempSeq.append(line);
  }
  temp.seq = new char[tempSeq.size()+1];
  strcpy(temp.seq,tempSeq.c_str());
  temp.bak = new char[tempSeq.size()+1];
  strcpy(temp.bak,tempSeq.c_str());
  temp.len = len;
  gaps.push_back(temp);

  temp.seq = new char[temp.len+1];
  strcpy(temp.seq,gaps[gaps.size()-1].seq);
  reverse_complement(temp.seq);
  gaps.push_back(temp);
  len = 0;
  tempSeq.clear();
  fclose(fptr);
}





void read_contigs(char* fileName){
  char line[10000];
  contigNode temp;
  string tempSeq;
  int len;
  char *ch;
  FILE *fptr;

   contigs.push_back(temp);  // dummy node for contig0
   fptr = fopen(fileName,"r");


  len = 0;
  tempSeq.clear();
  while(fgets(line,sizeof(line)-1,fptr)!=NULL){
         if(line[0]=='>' && len>0){
           temp.seq = new char[tempSeq.size()+1];
           strcpy(temp.seq,tempSeq.c_str());
           temp.bak = new char[tempSeq.size()+1];
           strcpy(temp.bak,tempSeq.c_str());
           temp.len = len;
           contigs.push_back(temp);
           len = 0;
           tempSeq.clear();
         }
         if(line[0]=='>'){
           len = 0;
           tempSeq.clear();
           continue;
         }
         ch = strchr(line,'\n');
         if(ch)
            *ch = '\0';
         len += strlen(line);
         tempSeq.append(line);
  }
  temp.seq = new char[tempSeq.size()+1];
  strcpy(temp.seq,tempSeq.c_str());
  temp.bak = new char[tempSeq.size()+1];
  strcpy(temp.bak,tempSeq.c_str());
  temp.len = len;
  contigs.push_back(temp);
  len = 0;
  tempSeq.clear();
  fclose(fptr);
}





int main(int argc, char* argv[]){
   int shift,contigNo,start,end;
   char *temp;
   char line[1000];
   char *startPtr,*endPtr; 
   char orientation,bakch;
   bool closed,inChain;
   int  gapID,index;
   FILE *fptr;
   bool *used_contig;




  if(argc!= 4){
    printf("./JRgapcloser contigFIle gap.fa *.evidence\n");
    return 0;
  }
  read_contigs(argv[1]);
  read_gaps(argv[2]); 


  used_contig = new bool[contigs.size()];
  memset(used_contig,0,sizeof(bool)*contigs.size());
  fptr = fopen(argv[3],"r");
  inChain = false;
  index = 1;
  while(fgets(line,sizeof(line)-1,fptr)!=NULL){
     if(line[0]=='>'){
       fgets(line,sizeof(line),fptr);
       sscanf(line,"%c_tig%d|",&orientation,&start);
       if(orientation == 'r')
          reverse_complement(contigs[start].seq);
       continue;
     }
     if(line[0]=='\n'){
        if(!inChain)
           printf(">contig%d_start%d\n",index++,start);
       printf("%s\n",contigs[start].seq);
       used_contig[start]=true;
      inChain = false;
      continue;
     }
    
     sscanf(line,"%c_tig%d|",&orientation,&end);
     if(orientation == 'r')
       reverse_complement(contigs[end].seq);
     closed = false;




     if((startPtr=strstr( contigs[end].seq, contigs[start].seq+contigs[start].len-MATCHLEN))){
         endPtr = contigs[end].seq;
         closed = true;

     }

     for( gapID=0;!closed && gapID<gaps.size();gapID++){
         bakch = contigs[end].seq[MATCHLEN];
         contigs[end].seq[MATCHLEN]='\0';
         if((startPtr=strstr(gaps[gapID].seq,contigs[start].seq+contigs[start].len-MATCHLEN)) && (endPtr=strstr(gaps[gapID].seq,contigs[end].seq))){
           contigs[end].seq[MATCHLEN] = bakch;
           closed = true;
           break;
         }
         contigs[end].seq[MATCHLEN]= bakch;
     }
     if(closed){
       if(!inChain)
          printf(">contig%d_start%d\n",index++,start);
       if(endPtr > startPtr){
          contigs[start].seq[contigs[start].len-MATCHLEN] = '\0';
          *endPtr = '\0';
          printf("%s%s",contigs[start].seq,startPtr);
       }else{
          int overlap = startPtr - endPtr;
          contigs[start].seq[contigs[start].len - MATCHLEN-overlap] = '\0';
          printf("%s",contigs[start].seq);
       }
      inChain = true;
     }else{
       if(!inChain)
         printf(">contig%d_start%d\n",index++,start);
       printf("%s\n",contigs[start].seq);
       inChain = false;        
     }
     used_contig[start]=true;
     start = end;
  }


   if(!used_contig[start]){
     if(closed){  // for the last contig
       if(!inChain)
          printf(">contig%d_start%d\n",index++,start);
       if(endPtr > startPtr){
          contigs[start].seq[contigs[start].len-MATCHLEN] = '\0';
          *endPtr = '\0';
          printf("%s%s",contigs[start].seq,startPtr);
       }else{
          int overlap = startPtr - endPtr;
          contigs[start].seq[contigs[start].len - MATCHLEN-overlap] = '\0';
          printf("%s",contigs[start].seq);
       }
      inChain = true;
     }else{
       if(!inChain)
         printf(">contig%d_start%d\n",index++,start);
       printf("%s\n",contigs[start].seq);
       inChain = false;
     }
     used_contig[start]=true;
   }

    for(int i=1;i<contigs.size();i++){
       if(!used_contig[i])
         printf(">contig%d\n%s\n",i,contigs[i].seq);
    }


  fclose(fptr);
  delete[] used_contig;
  return 0;
}
