#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <string>
#include <vector>
#include <math.h>
#include <algorithm>
#include <time.h>
#include <set>

using namespace std;

typedef struct contigNode{
 //   char name[1000];
    int len;
    int tailRead;
    int tailShift;
    int RCtailRead;
    int RCtailShift;
    char *seq;
    char *bak;
}contigNode;
typedef struct linkingNode{
   int id;
   int len;
}linkingNode;

typedef struct neuronNode{
         int index;
         int inverseID;
         int readNum;
         char *seq;
         bool used;
         int contigNo;
         bool visited;
         bool builded;
         set<linkingNode> output;
}neuronNode;



typedef struct anchorNode{
     int readID;
     int contigID;
}anchorNode;


vector<contigNode> contigs;

int seqNum,seqLen;
neuronNode *table;
int MINMATCH = 30;
int MAXMATCH;
int MINKMER = 0;
int MAXSTEP = 30;

const char ACGT[4] = {'A','C','G','T'};


void strLastN(char *s, char* t,int n){
        int len;
        int i,j;

        len = strlen(s);
        for(j=0,i=len-n;i<len;i++,j++)
           t[j] = s[i];
        t[j] = '\0';

}
 bool operator<(const linkingNode& left, const linkingNode& right) {
         return left.id < right.id;
 }

int compare(const void * a, const void * b){


   char *ptra,*ptrb;

   ptra = ((neuronNode*)a)->seq;
   ptrb = ((neuronNode*)b)->seq;
   while(*ptra !='\0' && *ptrb!='\0'){
        if(*ptra != *ptrb)
            return *ptra - *ptrb;
        ptra++;ptrb++;

   }
   return 0;
}

char complement(char ch){

     if(ch == 'A')
        return 'T';
     if(ch == 'C')
        return 'G';
     if(ch == 'G')
        return 'C';
     if(ch == 'T')
        return 'A';
    if(ch == 'a')
        return 't';
     if(ch == 'c')
        return 'g';
     if(ch == 'g')
        return 'c';
     if(ch == 't')
        return 'a';
     return ch;
}

 int ACGT2int(char ch){

    if(ch == 'A' || ch =='a')
      return 0;
    if(ch == 'C' || ch =='c')
      return 1;
    if(ch == 'G' || ch =='g')
      return 2;
    if(ch == 'T' || ch =='t')
      return 3;
    return 4;
}


void reverse_complement(char *str){
   int i,j,len;
  // char *temp;
   char bak;

   len = strlen(str);
   for(i=0,j=len-1;i<=j;i++,j--){
     bak = str[i];
     str[i] = complement(str[j]);
     str[j] = complement(bak);
   }
}

int read_seq(char *fileName){
   FILE *fptr;
   int i,n;


   fptr = fopen(fileName,"r");
   fscanf(fptr,"readNum= %d\treadLength= %d\n",&n,&seqLen);
   MAXMATCH = seqLen-1;
   seqLen++;
   table = new neuronNode[n];
   for(i=0;i<n;i++){
        table[i].seq = new char[seqLen];
        table[i].index = i;
        table[i].contigNo = -1;
        table[i].visited = false;
        fscanf(fptr,"%d\t%s\t%d\n",&(table[i].readNum),table[i].seq,&(table[i].inverseID));
  //      table[i].used = false;
  //      table[i].builded = false;
  //      table[i].seeded = false;
  //      readTotal += table[i].readNum;
   }
   fclose(fptr);
   return n;
}

void read_contigs(char* fileName){
  char line[10000];
  contigNode temp;
  string tempSeq;
  int len;
  char *ch;
  FILE *fptr;

   contigs.push_back(temp);  // dummy node for contig0
   fptr = fopen(fileName,"r");


  len = 0;
  tempSeq.clear();
  while(fgets(line,sizeof(line)-1,fptr)!=NULL){
         if(line[0]=='>' && len>0){
           temp.seq = new char[tempSeq.size()+1];
           strcpy(temp.seq,tempSeq.c_str());
           temp.bak = new char[tempSeq.size()+1];
           strcpy(temp.bak,tempSeq.c_str());
           contigs.push_back(temp);
           len = 0;
           tempSeq.clear();
         }
         if(line[0]=='>'){
           len = 0;
           tempSeq.clear();
           continue;
         }
         ch = strchr(line,'\n');
         if(ch)
            *ch = '\0';

         len += strlen(line);
         tempSeq.append(line);
  }
  temp.seq = new char[tempSeq.size()+1];
  strcpy(temp.seq,tempSeq.c_str());
  temp.bak = new char[tempSeq.size()+1];
  strcpy(temp.bak,tempSeq.c_str());
  contigs.push_back(temp);
  len = 0;
  tempSeq.clear();

  fclose(fptr);
}

void read_contigs_bak(char* fileName){
   contigNode temp;
   FILE *fptr;


   contigs.push_back(temp);  // dummy node for contig0
   fptr = fopen(fileName,"r");
   while(fscanf(fptr,">contig%*d_%d_%*s\n",&temp.len)!=EOF){
        temp.seq = new char[temp.len+1];
        fscanf(fptr,"%s\n",temp.seq);
        temp.bak = new char[temp.len+1];
        strcpy(temp.bak,temp.seq);
        contigs.push_back(temp);
   }
   fclose(fptr);
}

/*
void mark_used(char *seq, int contigNo){
    int i,last,inverseID;
    neuronNode temp;
    neuronNode *ptr;



    last = strlen(seq) - (seqLen-1);
    temp.seq = new char[seqLen];
    for(i = 0;i<=last;i++){
       strncpy(temp.seq,seq+i,(seqLen-1));
       temp.seq[seqLen-1]='\0';
       ptr = (neuronNode *) bsearch(&temp,&table[0],seqNum,sizeof(neuronNode),compare);

       if(ptr && !ptr->used){
          ptr->contigNo = contigNo;
          ptr->used = true;
          inverseID = ptr->inverseID;  //getInverseID(ptr->index);
          table[inverseID].used = true;
          table[inverseID].contigNo = -contigNo;
       }else if(ptr && !ptr->used)
         printf("multiple use.\n");
    }
    delete[] temp.seq;

}
*/
void mark_end_used(char *seq, int contigNo){
    int i,last,inverseID;
    neuronNode temp;
    neuronNode *ptr;
    int endSize = 50;


    last = strlen(seq) - (seqLen-1);
    temp.seq = new char[seqLen];
    for(i = 0;i<=endSize;i++){
       strncpy(temp.seq,seq+i,(seqLen-1));
       temp.seq[seqLen-1]='\0';
       ptr = (neuronNode *) bsearch(&temp,&table[0],seqNum,sizeof(neuronNode),compare);

       if(ptr && !ptr->used){
          ptr->contigNo = contigNo;
          ptr->used = true;
          inverseID = ptr->inverseID;  //getInverseID(ptr->index);
          table[inverseID].used = true;
          table[inverseID].contigNo = -contigNo;
       }else if(ptr && ptr->used){
//         printf("multiple use.\n");
          ptr->contigNo = 0;
          table[ptr->inverseID].contigNo = 0;
      }
    }

    for(i=strlen(seq)-endSize;i<=last;i++){
       strncpy(temp.seq,seq+i,(seqLen-1));
       temp.seq[seqLen-1]='\0';
       ptr = (neuronNode *) bsearch(&temp,&table[0],seqNum,sizeof(neuronNode),compare);
       if(ptr && !ptr->used){
          ptr->contigNo = contigNo;
          ptr->used = true;
          inverseID = ptr->inverseID;  //getInverseID(ptr->index);
          table[inverseID].used = true;
          table[inverseID].contigNo = -contigNo;
       }else if(ptr && ptr->used){
 //        printf("multiple use.\n");
          ptr->contigNo = 0;
          table[ptr->inverseID].contigNo = 0;

       }
    }
    delete[] temp.seq;

}

/*
int kmerExist(char *seq , double *ratio, int *gapLen,bool allowUsed, kmerNode* kmerpath){
    int i,j,last,n,preIndex,dist;
    char bakch;
    neuronNode temp;
    neuronNode *ptr;
    kmerNode *kmerIndex;


    last = strlen(seq) - (seqLen-1);
    preIndex = *gapLen = 0;
    kmerIndex = kmerpath;
    for(n = 0,i = 1,j=i+seqLen-1;i<last;i++,j++){
       if(kmerIndex == NULL){
          bakch = seq[j];
          seq[j] = '\0';
          temp.seq = seq+i;
          ptr = (neuronNode *) bsearch(&temp,&table[0],seqNum,sizeof(neuronNode),compare);
          seq[j] = bakch;
          if(ptr && (allowUsed || !ptr->used)){
             dist = i -preIndex;
             if(dist > *gapLen)
                *gapLen = dist;
             preIndex = i;
             n++;
          }
       }else{
          bakch = seq[j];
          if(kmerIndex->shift[ACGT2int(bakch)]){
             kmerIndex = kmerIndex->shift[ACGT2int(bakch)];
             n = (kmerIndex->exist)? n+1: n;
          }else{
             kmerIndex->shift[ACGT2int(bakch)] = new kmerNode;
             kmerIndex =  kmerIndex->shift[ACGT2int(bakch)];
             for(int k=0; k<4 ; k++)
                kmerIndex->shift[k] = NULL;
             bakch = seq[j];
             seq[j] = '\0';
             temp.seq = seq+i;
             ptr = (neuronNode *) bsearch(&temp,&table[0],seqNum,sizeof(neuronNode),compare);
             seq[j] = bakch;
             if(ptr && (allowUsed || !ptr->used)){
                kmerIndex->exist = true;
                n++;
             }else
                kmerIndex->exist = false;

          }
       }
    }


    *ratio = (double)n / last;
    return n;
}

*/
void buildForwardConnection(int i){
   int start,len;

//    double ratio;
    neuronNode *ptrOutput;
    neuronNode temp;
    char *concate;
    linkingNode templink;
    set<linkingNode>::iterator it,del;

      table[i].builded = true;
      temp.seq = new char[seqLen];
      concate = new char[2*seqLen];

      for(len=MINMATCH;len<=MAXMATCH;len++){
          // exactly match
          strLastN(table[i].seq,temp.seq,len);
          ptrOutput = (neuronNode *) bsearch(&temp,&table[0],seqNum,sizeof(neuronNode),compare);
          if(ptrOutput){
            for(start=ptrOutput->index;start > 0 && strncmp(table[start].seq,temp.seq,len)==0;start--){    // all read before pItem that starts with temp.seq
    //              if(table[start].used)
    //                 continue;
    //              sprintf(concate,"%s%s\0",table[i].seq,table[start].seq+len);
    //              concate[strlen(concate)-5] = '\0';
    //              if(kmerExist(concate+5,&ratio,&gapLen,false) > MINKMER && start != i ){
                  if(start !=i){
                       templink.id = start;
                       templink.len = len;
                       table[i].output.insert(templink);
                   }
  //                }
            }
            for(start=ptrOutput->index+1;start<seqNum && strncmp(table[start].seq,temp.seq,len)==0;start++){   // all reads after pItem that starts with temp.seq
   //               if(table[start].used)
   //                  continue;
   //               sprintf(concate,"%s%s\0",table[i].seq,table[start].seq+len);
  //                concate[strlen(concate)-5] = '\0';
  //                if(kmerExist(concate+5,&ratio,&gapLen,false) > MINKMER && start != i ){
                  if(start!=i){
                       templink.id = start;
                       templink.len = len;
                       table[i].output.insert(templink);
                   }
  //                }
            }
          }
      }
      delete[] temp.seq;
      delete[] concate;
}


int tail_read(char *seq, int *shift){
  bool backcheck;
  neuronNode temp;
  int last,templen;
  neuronNode *ptr = NULL;
  char tempch;

//          for(backcheck=false,last =strlen(pathit->seq)-1;last > seqLen;last--){
        temp.seq = new char[seqLen+1];
        *shift = 0;  
        for(backcheck=false,templen = strlen(seq)-40,last =strlen(seq);last > templen;last--,(*shift)++){
             tempch = seq[last];
             seq[last] = '\0';
             strLastN(seq, temp.seq, seqLen-1);
             ptr = (neuronNode *) bsearch(&temp,&table[0],seqNum,sizeof(neuronNode),compare);
             if(ptr){ // && travel(ptr->index,1,NULL,MINDENSITY)){
                 backcheck = true;
                 seq[last]='\0';
                 break;
             }
          }
       delete[] temp.seq;

       if(backcheck){
         return ptr->index;
       }
       return -1;
}


int RCtail_read(char *seq,int *shift){
    int i,last;
    neuronNode temp;
    neuronNode *ptr;
    char tempch;

    
    temp.seq = new char[seqLen];
    *shift = 0;
    for(i=0,last=seqLen-1;;i++,last++,(*shift)++){
       tempch = seq[last];
       seq[last] = '\0';
       strcpy(temp.seq,seq+i);
       reverse_complement(temp.seq);
       ptr = (neuronNode *) bsearch(&temp,&table[0],seqNum,sizeof(neuronNode),compare);
       seq[last]=tempch;
       if(ptr){ // && travel(ptr->index,1,NULL,MINDENSITY)){
            delete[] temp.seq;
            return ptr->index;
       }
    }
    delete[] temp.seq;
}

/*
void travel(int start, int contigNo,int step){

    set<linkingNode>::iterator itS;

   if(step > 3)
     return;

 // printf("%d %s\n",step,table[start].seq);
   buildForwardConnection(start);
   for(itS=table[start].output.begin();itS!=table[start].output.end();itS++){
      if(table[itS->id].used && !connected[contigNo][table[itS->id].contigNo]){
         connected[contigNo][table[itS->id].contigNo] = true;
        printf("%d %d\n",contigNo,table[itS->id].contigNo);
      }else if(!table[itS->id].visited){

         travel(itS->id,contigNo,step+1);
      }
   }
}
*/


void BFS_travel(int start,int contigNo,set<int> *connected,vector<anchorNode> *anchor){
    set<linkingNode>::iterator itS;
    vector<int> list;
    int temp;
    int begin,last,step;
    anchorNode tempAnchor;
    bool unique;


     temp= start;
     list.push_back(temp);
     begin = 0;
     last = 1;
     unique = true;
     for(step=1;unique && step<MAXSTEP;step++){
       for(;unique && begin<last;begin++){
         buildForwardConnection(list[begin]);
         for(itS=table[list[begin]].output.begin();itS!=table[list[begin]].output.end();itS++){
            if(   table[itS->id].used && table[itS->id].contigNo !=0 ){
                connected->insert(table[itS->id].contigNo);
                tempAnchor.contigID = table[itS->id].contigNo;
                tempAnchor.readID = itS->id;
                anchor->push_back(tempAnchor);
     //           if(connected->size()>1){
     //             unique = false;
     //             break;
     //           } 
            }else if(!table[itS->id].visited){
               temp= itS->id;
               table[itS->id].visited = true;
               list.push_back(temp);
//printf("visit %s\n",table[temp].seq);
            }
         }
       }
       last = list.size();
     }

    for(int i=0;i<(int)list.size();i++){
       table[list[i]].visited = false;
    }
    list.clear();
}


bool *merged;

void merge_path(int i, vector<int> *path,set<int> *connected, set<int> *RCconnected){
    int start,target;
    bool RC ;
    bool test;


     start = i;
     RC = true;
     while(test = (RC?RCconnected[start].size()==1:connected[start].size()==1)){
        target = (RC)? *(RCconnected[start].begin()):*(connected[start].begin());
        if(target > 0){
           if( !merged[target] && RCconnected[target].size()==1 && *(RCconnected[target].begin())==start){
                    path->push_back(-target);
                    merged[target] = true;
                    RC = false;
                    start = target;
           }else
                break;
        }else{
           if( !merged[-target] && connected[-target].size()==1 &&  *(connected[-target].begin())== start){
                   path->push_back(-target);
                   merged[-target]=true;
                   RC = true;
                   start = -target;
           }else
              break;
        }
     }
     std::reverse(path->begin(),path->end());
     path->push_back(i);
     start = i;
     RC = false;
     while(test = (RC?RCconnected[start].size()==1:connected[start].size()==1)){
        target = (RC)?*(RCconnected[start].begin()):*(connected[start].begin());
          if(target > 0){
              if( !merged[target] && RCconnected[target].size()==1 && *(RCconnected[target].begin())==-start){
                   path->push_back(target);
                   merged[target]=true;
                   start = target;
                   RC = false;
              }else
                break;
          }else{
              if( !merged[-target] && connected[-target].size()==1 && *( connected[-target].begin()) == -start){
                   path->push_back(target);
                   merged[-target]=true;
                   start = -target;
                   RC = true;
              }else
                 break;
          }
     }
}


typedef struct gapNode{
    int id;
    char seq[5000];
}gapNode;

void showGap(int start,int end, char* gap){

    set<linkingNode>::iterator itS;
    vector<gapNode> list;
    gapNode temp;
    int begin,last,step;
    bool found;

     temp.id= start;
     memset(temp.seq,0,sizeof(temp.seq));
     list.push_back(temp);
     begin = 0;
     last = 1;
     found = false;
     for(step=1;step<MAXSTEP && !found;step++){
       for(;begin<last && !found ;begin++){
         for(itS=table[list[begin].id].output.begin();itS!=table[list[begin].id].output.end();itS++){
            if( itS->id == end ){
              strcpy(gap,list[begin].seq);
              strcat(gap,table[end].seq+(itS->len));
              found = true;
              break;
            }else if(!table[itS->id].visited){
               temp.id = itS->id;
               strcpy(temp.seq,list[begin].seq);
               strcat(temp.seq,table[itS->id].seq+itS->len);
               table[itS->id].visited = true;
               list.push_back(temp);
            }
         }
       }
       last = list.size();
     }

    for(int i=0;i<(int)list.size();i++){
       table[list[i].id].visited = false;
    }
    list.clear();

}
int main(int argc, char* argv[]){
   int shift,contigNo;
   char *temp;
   set<linkingNode>::iterator itS,itT,Send,Tend;
   char gap[5000];
   set<int> *connected;
   set<int> *RCconnected; 
   vector<anchorNode> *anchor,*RCanchor;
   vector<int> path;
   char *chptr; 

  if(argc!= 4){
    printf("final contigFIle hisFile MINOVERLAP.\n");
    return 0;

  }
  read_contigs(argv[1]);
  seqNum = read_seq(argv[2]);
  MINMATCH = atoi(argv[3]);

  for(int i=1; i< (int)contigs.size();i++)
    mark_end_used(contigs[i].seq,i);

  
  connected = new set<int>[contigs.size()];
  RCconnected = new set<int>[contigs.size()];
  anchor = new vector<anchorNode>[contigs.size()];
  RCanchor = new vector<anchorNode>[contigs.size()];





  for(int i=1; i< (int)contigs.size();i++){
    contigs[i].tailRead = tail_read(contigs[i].seq,&shift);
 //   contigs[i].seq[strlen(contigs[i].seq)-shift] = '\0';
//printf(">contig%d %d ",i,shift);
    contigs[i].RCtailRead = RCtail_read(contigs[i].seq,&shift);
//printf("%d\n",shift);
    if(shift != 0){
      temp = new char[strlen(contigs[i].seq)+1];
      strcpy(temp,contigs[i].seq+shift);
      delete[] contigs[i].seq;
      contigs[i].seq = temp;
    }

    if(contigs[i].tailRead > 0){
        BFS_travel(contigs[i].tailRead,i,&connected[i],&anchor[i]);
    }

    if(contigs[i].RCtailRead > 0){
        BFS_travel(contigs[i].RCtailRead,i,&RCconnected[i],&RCanchor[i]);
    }

  }


/*
for(int i=1;i<contigs.size();i++){
   printf("contig%d\n",i);

   if(connected[i].size()>0){
     printf(" connect:");
     for(set<int>::iterator it=connected[i].begin();it!=connected[i].end();it++)
          printf("%d\t",*it);
      printf("\n");
   }
   if(RCconnected[i].size()>0){
    printf(" RCconnect:");
    for(set<int>::iterator it=RCconnected[i].begin();it!=RCconnected[i].end();it++)
        printf("%d\t",*it);
   }
   printf("\n");
   
}
*/
   merged = new bool[contigs.size()];
   memset(merged,0,sizeof(bool)*contigs.size());
   contigNo = 1;
   for(int i=1; i<(int)contigs.size();i++){
       if(merged[i])
          continue;
       path.clear();
       merged[i] = true;
       merge_path(i,&path,connected,RCconnected);
       if(path.size()==1){
            printf(">contig%d\n%s\n",contigNo++,contigs[i].bak);

       }else{
         printf(">contig%d_",contigNo++);
         for(int j=0;j<(int)path.size();j++)
             printf("%d_",path[j]);
         printf("\n");

         if(path[0]>0)
              printf("%s",contigs[path[0]].seq); 
         else{
             reverse_complement(contigs[-path[0]].seq);
             printf("%s",contigs[-path[0]].seq);
         }

         for(int j=1;j<(int)path.size();j++){
            if(path[j-1]>0){
              for(int k=0;k<(int)anchor[path[j-1]].size();k++){
                 if(anchor[path[j-1]][k].contigID==path[j]){
                    memset(gap,0,sizeof(gap));
                    showGap(contigs[path[j-1]].tailRead,anchor[path[j-1]][k].readID,gap);
                    printf("%s",gap);
                    if(path[j] > 0){
                      chptr = strstr(contigs[path[j]].seq,table[anchor[path[j-1]][k].readID].seq);
                      if(chptr){
                        printf("%s",chptr+seqLen-1);
                      }else
                        printf("error");
                    }else{
                      reverse_complement(contigs[-path[j]].seq);
                      chptr = strstr(contigs[-path[j]].seq,table[anchor[path[j-1]][k].readID].seq);
                      if(chptr){
                        printf("%s",chptr+seqLen-1);
                      }else
                        printf("error");               
 
                    }
                    break;
                 }

              }
            }else{
              for(int k=0;k<(int)RCanchor[-path[j-1]].size();k++){
                   if(RCanchor[-path[j-1]][k].contigID==path[j]){
                     memset(gap,0,sizeof(gap));
                     showGap(contigs[-path[j-1]].RCtailRead,RCanchor[-path[j-1]][k].readID,gap);
                     printf("%s",gap);
                     if(path[j]>0){
                         chptr = strstr(contigs[path[j]].seq,table[RCanchor[-path[j-1]][k].readID].seq);
                      if(chptr){
                        printf("%s",chptr+seqLen-1);
                      }else
                        printf("error");
                     }else{
                        reverse_complement(contigs[-path[j]].seq);
                        chptr = strstr(contigs[-path[j]].seq,table[anchor[-path[j-1]][k].readID].seq);
                        if(chptr){
                          printf("%s",chptr+seqLen-1);
                        }else
                          printf("error");
                     }
                     break;
                 }

              }

            }
         }
         printf("\n");

       }

  }

  for(int i=0; i<(int)contigs.size();i++){
     connected[i].clear();
     RCconnected[i].clear();
     anchor[i].clear();
     RCanchor[i].clear();
  }
  delete[] connected;
  delete[] RCconnected;
  delete[] anchor;
  delete[] RCanchor;
  contigs.clear();
  delete[] merged;
  return 0;
}
