#include <stdio.h>
#include <string.h>
#include <string>
#include <vector>
#include <algorithm>
#include <math.h>
#include <iostream>
#define MAXCOUNT 1000
using namespace std;

typedef struct contigNode{
    char name[1000];
    string seq;
    int length;
    int *cov;
}contigNode;

vector<contigNode> contig;
int readLen;


bool compare(contigNode a, contigNode b){

        return  strcmp( a.name, b.name) < 0;

}
bool Lencompare(contigNode a, contigNode b){

        return   (a.length - b.length) > 0;

}


void readContig(FILE *fptr){
  char line[10000];
  contigNode temp;
  string tempSeq;
  int len;
  char *ch;

  len = 0;
  tempSeq.clear();
  while(fgets(line,sizeof(line)-1,fptr)!=NULL){
         if(line[0]=='>' && len>0){
           temp.length = len;
           temp.seq = tempSeq;
           contig.push_back(temp);
           len = 0;
           tempSeq.clear();
         }
         if(line[0]=='>'){
//           sscanf(line,">%s",&(temp.name));
           strcpy(temp.name,line);
           len = 0;
           tempSeq.clear();
           continue;      
         }    
         ch = strchr(line,'\n');
         if(ch)
            *ch = '\0';
         
         len += strlen(line);
         tempSeq.append(line);
  }
  temp.length = len;
  temp.seq = tempSeq;
  contig.push_back(temp);
/*
  sort(contig.begin(),contig.end(),compare);

   for(int i=0; i< contig.size();i++){
     contig[i].cov = new int[contig[i].length+1];
     memset(contig[i].cov,0,sizeof(int)*(contig[i].length+1));
   }
*/
}

int SearchCompare(const void * a, const void * b){

        return  strcmp( ((contigNode*)a)->name, ((contigNode*)b)->name);

}


void smooth(int windowSize){
   int i,j,sum;
   

   for(i=0;i<contig.size();i++){
    for(sum=j=0;j<windowSize;j++)
       sum+= contig[i].cov[j];
    for(j=0;j+windowSize<contig[i].length;j++)
       contig[i].cov[j] = sum / windowSize;
       sum -= contig[i].cov[j];
       sum += contig[i].cov[j+windowSize];
   }
}

void readCov(FILE *fhit){
  int len,pos;
  char strand;  
  char line[100000],name[1000];
  contigNode temp;
  contigNode *ptr;

  while(fgets(line,sizeof(line),fhit)!=NULL){
     if(line[0]=='\t')
       sscanf(line,"\t%*s\t%*s\t%*d\t%*c\t%d\t%c\t%s\t%d\t",&len,&strand,&name,&pos);
     else
       sscanf(line,"%*s\t%*s\t%*s\t%*d\t%*c\t%d\t%c\t%s\t%d\t",&len,&strand,&name,&pos);
     strcpy(temp.name,name);
     ptr = (contigNode *) bsearch(&temp,&contig[0],contig.size(),sizeof(contigNode),SearchCompare);
     if(ptr){
       for(int i = pos; i< pos+len;i++)
         ptr->cov[i]++;     
     }
  }
  readLen = len;
//  smooth(readLen);

}

void calCov(int minLength,double *avgCov, double *std){
   int i,j,n;
   long long int totalCov_2,totalCov;
   bool inLowCov;

   totalCov = totalCov_2 = n = 0;

   for(i=0;i<contig.size();i++){
     if(contig[i].length < minLength)
         break;
     for(j=readLen+1;j<contig[i].length-readLen;j++){
        totalCov_2 += contig[i].cov[j]*contig[i].cov[j];
        totalCov += contig[i].cov[j];
        n++;
     }
   }

  printf("total Cov = %d\n",totalCov);
  *avgCov = (double) totalCov / n;
  *std = sqrt(totalCov_2 /n - (*avgCov)* (*avgCov));
  printf("n = %d\n",n);
//  printf("total Cov square = %lld\n",totalCov_2);
  printf("average Coverage All = %lf\n",*avgCov);
  printf("std All = %lf\n",*std);
}

void calLowCov(int minLength, double avgCov, double std){
   int i,j,k,lowLen,lowCount;
   int threshold;
   bool inLowCov;
   double windowCov;
   
   threshold = (int)(avgCov - std);
   lowCount = lowLen = 0;
   inLowCov = false;
   for(i=0;i<contig.size();i++){
     if(contig[i].length < minLength)
         break;
     for(j=readLen+1;j+readLen<contig[i].length-readLen;j++){
        windowCov = 0;
        for(k=j;k<j+readLen;k++)
          windowCov += contig[i].cov[k];
//        if(contig[i].cov[j] < threshold){
          if(windowCov /readLen < threshold){
             lowLen++;
             if(!inLowCov){
                lowCount++;
                inLowCov = true;
             }
        }else if(inLowCov)
           inLowCov = false;
     }
   }
   printf("threshold = %d\n",threshold);
   printf("Low Coverage Count = %d\n",lowCount);
   printf("Low Coverage Length = %d\n",lowLen);
}


int calN50(){
   int acc,i,median;
   int Sum;


   for(Sum=i=0;i<contig.size();i++)
       Sum += contig[i].length;
   median = Sum / 2;
   for(acc=i=0;i<contig.size();i++){
      acc += contig[i].length;
      if(acc >= median)
        return contig[i].length;
   }

}

int histogram(double ratio){
  int i,j;
  int count[MAXCOUNT];
  int maxCov,totalSize,temp;

   totalSize = maxCov = 0;
   memset(count,0,sizeof(int)*MAXCOUNT);
   for(i=0;i<contig.size();i++){
      totalSize += contig[i].length;
      for(j=readLen+1;j+readLen<contig[i].length-readLen;j++){
        if(contig[i].cov[j]< MAXCOUNT){
           count[contig[i].cov[j]]++;
           if(contig[i].cov[j] > maxCov)
              maxCov = contig[i].cov[j];
        }else{
           count[MAXCOUNT-1]++;
           maxCov = MAXCOUNT;
        }
      }
   }
   for(i=0;i<maxCov;i++){
      temp += count[i];
      if(temp > totalSize * ratio)
         return i - 1;

   }
/*
   for(i=0;i<maxCov;i++){
     printf("%d\t%d\n",i,count[i]);
   }
*/
}

void outputLowCov(int minLength,int threshold){

   int i,j,k,lowLen,lowCount;
   bool inLowCov;
   double windowCov;

   printf("threshold = %d\n",threshold);
   for(i=0;i<contig.size();i++){
     if(contig[i].length < minLength)
         break;
     lowCount = lowLen = 0;
     inLowCov = false;
     for(j=readLen+1;j+readLen<contig[i].length-readLen;j++){
        windowCov = 0;
        for(k=j;k<j+readLen;k++)
          windowCov += contig[i].cov[k];
//        if(contig[i].cov[j] < threshold){
          if(windowCov /readLen < threshold){
             lowLen++;
             if(!inLowCov){
                lowCount++;
                inLowCov = true;
             }
        }else if(inLowCov)
           inLowCov = false;
     }
     printf("%s\t%d\t%d\n",contig[i].name,lowCount,contig[i].length);
//     printf("%s\t%d\t%d\n",contig[i].name,lowLen,contig[i].length);

   }
}

int main(int argc, char* argv[]){
   FILE *fcontig;
   int i,j,k,l,N50,minlength,maxlength;
   char *cstr;

   if(argc != 4){
     printf("contigLenFilter *.contigs minlength maxlength.\n");
     return 0;
   }
   fcontig = fopen(argv[1],"r");
   if(!fcontig){
      printf("%s open error.\n",argv[1]);
      return 0;
   }
   minlength = atoi(argv[2]);
   maxlength = atoi(argv[3]);
   readContig(fcontig);
   for(i=0; i< contig.size();i++){
      if(contig[i].length >= minlength && contig[i].length <=maxlength){
          strtok(contig[i].name,"\n");
          printf("%s_%d\n",contig[i].name,contig[i].length);
          cout<<contig[i].seq<<endl;
      }
   }

 //  sort(contig.begin(),contig.end(),Lencompare);
 //  N50 = calN50();

   for(i=0;i<contig.size();i++)
     contig[i].seq.clear();   

   contig.clear();
   fclose(fcontig);
   return 0;
}


